package de.unisaar.jfap.ex5.task3;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Provide method to find smallest directory to delete so that update can be applied.
 *
 * @author Jörg Steffen, DFKI
 */
public final class DirFinder {

  private static final Logger logger = LoggerFactory.getLogger(DirFinder.class);

  private static Pattern cdPattern = Pattern.compile("\\$ cd (.+)");
  private static Pattern fileItemPattern = Pattern.compile("(\\d+) (.+)");

  private DirFinder() {
    // private constructor to enforce noninstantiability
  }

  private static Map<String, Directory> initDirMap(List<String> lines) {

    // map path to the corresponding directory
    Map<String, Directory> dirMap = new LinkedHashMap<>();

    // init stack with top-level directory
    Stack<Directory> dirStack = new Stack<>();
    Directory topLevelDir = new Directory("");
    dirStack.push(topLevelDir);
    dirMap.put(topLevelDir.getPath(), topLevelDir);

    for (int i = 1; i < lines.size(); i++) {
      // ignore first line "cd /", as stack is already initialized with top-level directory
      String oneLine = lines.get(i);
      logger.debug(oneLine);

      // line with file item
      Matcher fileMatcher = fileItemPattern.matcher(oneLine);
      if (fileMatcher.matches()) {
        // add file to current directory, i.e. directory on top of stack
        dirStack.peek().addFile(new File(fileMatcher.group(2), fileMatcher.group(1)));
        continue;
      }

      // line with cd command
      Matcher cdMatcher = cdPattern.matcher(oneLine);
      if (cdMatcher.matches()) {
        String dirString = cdMatcher.group(1);
        if (dirString.equals("..")) {
          dirStack.pop();
        } else {
          Directory newDir = new Directory(dirStack.peek().getPath() + "/" + dirString);
          // add directory to current directory, i.e. directory on top of stack
          dirStack.peek().addDirectory(newDir);
          // new directory becomes current directory, i.e. top of stack
          dirStack.push(newDir);
          dirMap.put(newDir.getPath(), newDir);
        }
        continue;
      }

      // dir
      // we can ignore dir entries, as we handle the directory
      // when 'cd' for that directory is called
    }
    return dirMap;
  }

  /**
   * Find the smallest directory to delete so that update can be installed.
   *
   * @param path file to read console output from
   * @param diskSize the total disk size in bytes
   * @param updateSize the size of the update in bytes
   * @return the directory to delete
   * @throws IOException if reading console output fails
   */
  public static Directory findDirToDelete(Path path, int diskSize, int updateSize)
      throws IOException {

    List<String> lines = Files.readAllLines(path);
    Map<String, Directory> dirMap = initDirMap(lines);

    // disk usage is the size of the top directory /, stored under ""
    int diskUsage = dirMap.get("").getSize();
    int freeDiskSpace = diskSize - diskUsage;
    int toDelete = updateSize - freeDiskSpace;
    logger.info(String.format("total disk usage: %,d", dirMap.get("").getSize()));
    logger.info(String.format("free disk: %,d", freeDiskSpace));
    logger.info(String.format("to delete: %,d", toDelete));

    // get all directories and sort them by size (largest first)
    List<Directory> dirs = new ArrayList<>(dirMap.values());
    Collections.sort(dirs);
    Directory lastDir = null;
    for (Directory oneDir : dirs) {
      logger.debug("{}", oneDir);
      if (oneDir.getSize() < toDelete) {
        // as soon as we find a directory smaller than the required minimum size,
        // the previous directory is the one to delete
        break;
      }
      lastDir = oneDir;
    }
    logger.info("directory to delete: {}", lastDir);
    return lastDir;
  }
}
